// Hey Emacs! This is -*- c++ -*- ! :-)
// objcal.cc	-- Hans-Peter Stoerr Tue Dec 22 1998
// object oriented interface to cal-2.1 BDD library. Implementation.

#include <assert.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
// #include <algo>

#define HAVE_STDARG_H 1
#define HAVE_VARARGS_H 1
extern "C" {
#include <cal.h>
#include <calInt.h>
}

#include "objcal.h"
#include "calSupply.h"

// wo don't allow inlines while debugging.
#define inline
#include "objcal_inline.h"

// ================================================================
// global variable(s)

BddMgr mgr;
map<Bdd,int,Bdd::BddCompare>  Bdd::sSingleVarAssocIds; // Assocs for each single variable

// ================================================================
// misc. helper functions

#if !defined(linux)
extern "C" { long random(); }
#endif

static bool rndBool()
{
  long res=random();
  return ((res*(res/1<<16))>>16)%2;
}

static char lastchar(string s)
{
  return *(s.end()-1);
}

// ================================================================
// Operatoren

Bdd Bdd::operator+(const Bdd b) const
{
  assert(IsValid() && b.IsValid());
  return Cal_BddOr(mgr,rep,b);
}

Bdd Bdd::operator*(const Bdd b) const
{
  assert(IsValid() && b.IsValid());
  return Cal_BddAnd(mgr,rep,b);
}

Bdd Bdd::operator^(const Bdd b) const
{
  assert(IsValid() && b.IsValid());
  return Cal_BddXor(mgr,rep,b);
}

Bdd Bdd::operator-() const
{
  assert(IsValid());
  return Cal_BddNot(mgr,rep);
}

bool Bdd::operator==(const Bdd b) const
{
  assert(IsValid() && b.IsValid());
  return Cal_BddIsEqual(mgr,rep,b);
}

bool Bdd::operator!=(const Bdd b) const
{
  return !(*this==b);
}

// ================================================================
// tests

bool Bdd::IsValid() const
{
  return rep.IsValid();
}

bool Bdd::BddRep::IsValid() const
{
  return theBdd!=0 && Cal_BddType(mgr,theBdd)!=CAL_BDD_TYPE_OVERFLOW;
}

bool Bdd::IsVariable() const
{
  assert(IsValid());
  return Cal_BddType(mgr,rep)==CAL_BDD_TYPE_POSVAR;
}

bool Bdd::IsCube() const
{
  assert(IsValid());
  return Cal_BddIsCube(mgr,rep);
}

bool Bdd::IsConst() const
{
  assert(IsValid());
  return Cal_BddIsBddConst(mgr,rep);
}

bool Bdd::IsOne() const
{
  assert(IsValid());
  return Cal_BddIsBddOne(mgr,rep);
}

bool Bdd::IsZero() const
{
  assert(IsValid());
  return Cal_BddIsBddZero(mgr,rep);
}

bool Bdd::Intersects(Bdd p) const { // returns true if (*this)*p != mgr.Zero()
  return mgr.Zero()!=(*this)*p;
}

// ================================================================
// Misc Operations

Bdd Bdd::Equiv(Bdd b) const
{
  return Cal_BddXnor(mgr,rep,b);
}

Bdd ITE(const Bdd ifB,const Bdd thenB , const Bdd elseB) { // if then else operation
  return Cal_BddITE(mgr,ifB,thenB,elseB);
}

Bdd Between(const Bdd thisB, const Bdd thatB) { // minimal Bdd between thisB and thatB.
  return Cal_BddBetween(mgr,thisB,thatB);
}

Bdd Bdd::Constrain(const Bdd c) const { // generalized cofactor wrt. c
  return Cal_BddCofactor(mgr,rep,c);
}

Bdd Bdd::Reduce(const Bdd c) const { // returns poss. smaller Bdd which agrees on c
  if (mgr.Zero()==c) return mgr.One(); // work around Bug in cal-2.1
  if (mgr.One()==c) return *this;
  return Cal_BddReduce(mgr,rep,c);
}

Bdd MultiwayAnd(const BddVector& set) { // returns And of all Bdds of Set
  if (0==set.size()) return mgr.One();
  Cal_Bdd  *bdd_array=new Cal_Bdd[set.size()+1];
  Cal_Bdd  *dit=bdd_array;
  BddVector::const_iterator sit=set.begin();
  for(; set.end()!=sit; ++sit, ++dit) *dit=Cal_Bdd(*sit);
  *dit=NULL;
  Bdd res=Cal_BddMultiwayAnd(mgr,bdd_array);
  delete[] bdd_array;
  return res;
}

Bdd MultiwayOr(const BddVector& set) { // returns Or of all Bdds of Set
  if (0==set.size()) return mgr.Zero();
  Cal_Bdd  *bdd_array=new Cal_Bdd[set.size()+1];
  Cal_Bdd  *dit=bdd_array;
  BddVector::const_iterator sit=set.begin();
  for(; set.end()!=sit; ++sit, ++dit) *dit=Cal_Bdd(*sit);
  *dit=NULL;
  Bdd res=Cal_BddMultiwayOr(mgr,bdd_array);
  delete[] bdd_array;
  return res;
}

Bdd MultiwayXor(const BddVector& set) { // returns Xor of all Bdds of Set
  Cal_Bdd  *bdd_array=new Cal_Bdd[set.size()+1];
  Cal_Bdd  *dit=bdd_array;
  BddVector::const_iterator sit=set.begin();
  for(; set.end()!=sit; ++sit, ++dit) *dit=Cal_Bdd(*sit);
  *dit=NULL;
  Bdd res=Cal_BddMultiwayXor(mgr,bdd_array);
  delete[] bdd_array;
  return res;
}

long Size(BddVector& set)    // returns joint size of Bdds.
{
  Cal_Bdd  *bdd_array=new Cal_Bdd[set.size()+1];
  Cal_Bdd  *dit=bdd_array;
  BddVector::const_iterator sit=set.begin();
  for(; set.end()!=sit; ++sit, ++dit) *dit=Cal_Bdd(*sit);
  *dit=NULL;
  long res=Cal_BddSizeMultiple(mgr,bdd_array,1);
  delete[] bdd_array;
  return res;
}

// ================================================================
// Quantification
Bdd Bdd::Exists(const BddAssoc& assoc) const
{
  Cal_AssociationSetCurrent(mgr,assoc.Id());
  assert(!assoc.IsMapping());
  return Cal_BddExists(mgr,rep);
}

Bdd Bdd::Exists(Bdd b) const
{
  Cal_AssociationSetCurrent(mgr,MakeSingleVarAssocId(b));
  return Cal_BddExists(mgr,rep);
}

Bdd Bdd::Exists(const BddVector& v) const
{
  unsigned vsiz=v.size();
  if (vsiz==0) return *this;
  if (vsiz==1) return this->Exists(v[0]);
  if (vsiz<8) {
    // stupidly, this seems to be faster than the process of converting the vector to an Association
    // and doing Exists.
    BddVector::const_iterator it=v.begin(), end=v.end();
    Bdd res=*this;
    for (;end!=it;++it) res=res.Exists(*it);
    return res;
  } else {
    BddAssoc      ass(v);
    return this->Exists(ass);
  }
}

Bdd Bdd::ForAll(const BddAssoc& assoc) const
{
  Cal_AssociationSetCurrent(mgr,assoc.Id());
  assert(!assoc.IsMapping());
  return Cal_BddForAll(mgr,rep);
}

Bdd Bdd::ForAll(Bdd b) const
{
  Cal_AssociationSetCurrent(mgr,MakeSingleVarAssocId(b));
  return Cal_BddForAll(mgr,rep);
}

Bdd Bdd::ForAll(const BddVector& v) const
{
  BddAssoc      ass(v);
  return this->ForAll(ass);
}

Bdd Bdd::RelProd(const BddAssoc& assoc, Bdd b) const
{
  Cal_AssociationSetCurrent(mgr,assoc.Id());
  assert(!assoc.IsMapping());
  return Cal_BddRelProd(mgr,rep,b);
}

Bdd Bdd::RelProd(Bdd e, Bdd b) const
{
  Cal_AssociationSetCurrent(mgr,MakeSingleVarAssocId(e));
  return Cal_BddRelProd(mgr,rep,b);
}

int Bdd::MakeSingleVarAssocId(Bdd b) const
{
  map<Bdd,int,BddCompare>::const_iterator  it=sSingleVarAssocIds.find(b);
  if (sSingleVarAssocIds.end()!=it) return it->second;
  // generate new Assoc
  assert(b.IsVariable());
  Cal_Bdd assInfo[2]={b.rep,NULL};
  int assocId=Cal_AssociationInit(mgr,assInfo,0);
  assert(assocId!=-1);
  sSingleVarAssocIds[b]=assocId;
  return assocId;
}

// ================================================================
// Substitution
Bdd Bdd::Substitute(const BddAssoc& assoc) const
{
  Cal_AssociationSetCurrent(mgr,assoc.Id());
  assert(assoc.IsMapping());
  if(assoc.MapsToVariables()) return Cal_BddVarSubstitute(mgr,rep);
  else return Cal_BddSubstitute(mgr,rep);
}

// ================================================================
// Set of Support etc.

BddVector Bdd::Support() const
  // Set of Support.
{
  BddVector     res;
  Cal_Bdd       support[Cal_BddVars(mgr)+1];
  Cal_BddSupport(mgr,rep,support);
  for(int i=0; NULL!=support[i]; i++)
    res.push_back(support[i]);
  return res;
}

bool Bdd::DependsOn(const Bdd var) const
  // does *this depend on var?
{
  return Cal_BddDependsOn(mgr,rep,var);
}

Bdd Bdd::Satisfy() const
{
  return Cal_BddSatisfy(mgr,rep);
}

Bdd Bdd::SatisfySupport(const BddAssoc& assoc) const
{
  assert(!assoc.IsMapping());
  Cal_AssociationSetCurrent(mgr,assoc.Id());
  return Cal_BddSatisfySupport(mgr,rep);
}

Bdd Bdd::RandomPath() const
{
  assert(*this!=mgr.Zero());
  if(*this==mgr.One()) return *this;
  if(rndBool()) {
    Bdd thp=ThenPart();
    if(thp!=mgr.Zero())
      return IfPart()*thp.RandomPath();
    else 
      return -IfPart()*ElsePart().RandomPath();
  } else {
    Bdd elp=ElsePart();
    if(elp!=mgr.Zero())
      return -IfPart()*elp.RandomPath();
    else 
      return IfPart()*ThenPart().RandomPath();
  }
}

// ================================================================

Bdd Bdd::SuperCube() const
  // returns smallest cube larger than current Bdd, but with (at most) the same vars
  // the implementation is a bit inefficient, but I don't know any better now.
{
  CSupVarCheckResT      varTabl[Cal_BddVars(mgr)+1];
  CSupBddVariableCheck(mgr,rep,varTabl);
  BddVector     res; Bdd var;
  int i=0;
  while (NULL!=varTabl[i].bddVar) {
    switch(varTabl[i].dep) {
    case POSDEP: var=Bdd(varTabl[i].bddVar);
      if(mgr.One()==this->Implies(var)) res.push_back(var); break;
    case NEGDEP: var=-Bdd(varTabl[i].bddVar);
      if(mgr.One()==this->Implies(var)) res.push_back(var); break;
    case BOTHDEP: break;        // nothing to do here
    case INDEP: assert(0); abort(); // this shouldn't happen.
    default: assert(0); abort(); // this shouldn't happen as well.
    }
    i++;
  }
  if (0==res.size()) return mgr.One();
  Bdd   theResult=MultiwayAnd(res);
  assert(mgr.One()==this->Implies(theResult));
  return theResult;
}

// ================================================================
// misc.

string Bdd::VarName() const
{
  assert(IsVariable());
  return mgr.varName(*this);
}

char* BddMgr::printNameVars(Cal_BddManager, Cal_Bdd varpt, Cal_Pointer_t)
{
  Bdd var(varpt,Bdd::unfree);
  return (char *)var.VarName().c_str();
}

void Bdd::Print(const char* name) const
{
  if(name) printf("%s = ",name);
  if(*this==mgr.One())
    printf("true");
  else if(*this==mgr.Zero())
    printf("false");
  else {
    string res;
    Bdd::printCubes(*this,&res,"",false,10000);
    printf(res.c_str());
  }
  printf("\n");
}

bool Bdd::printCubes(Bdd b, string* res, string start, bool plus, int maxlen)
{
  if (unsigned(maxlen)<res->size()) { *res+="..."; return plus; }
  if (b==mgr.One()){      // simple cases
    if(plus) *res+=" + ";
    *res+=start;
    return true;
  } else if(b==mgr.Zero()) return plus;
  // we have a nonterminal Bdd
  string ifName=b.IfPart().VarName();
  Bdd thenpart=b.ThenPart(), elsepart=b.ElsePart();
  if(start!="") start+="*";
  plus=printCubes(thenpart,res,start+ifName,plus,maxlen);
  if (unsigned(maxlen)<res->size()) { *res+="..."; return plus; }
  plus=printCubes(elsepart,res,start+"-"+ifName,plus,maxlen);
  return plus;
}

string Bdd::Describe(int maxlen) const
{
  if(*this==mgr.One())
    return "true";
  else if(*this==mgr.Zero())
    return "false";
  string res;
  Bdd::describe(*this,&res,"",false,maxlen);
  return res;
}

bool Bdd::describe(Bdd b, string *res, string start, bool plus,int maxlen)
{
  if (unsigned(maxlen)<res->size()) { *res+="..."; return plus; }
  if(b==mgr.One()){      // simple cases
    if(plus) *res += " + ";
    *res += start;
    return true;
  } else if(b==mgr.Zero()) return plus;
  // we have a nonterminal Bdd
  string ifName=b.IfPart().VarName();
  Bdd thenpart=b.ThenPart(), elsepart=b.ElsePart();
  if(start!="") start+="*";
  plus=printCubes(thenpart,res,start+ifName,plus,maxlen);
  if (unsigned(maxlen)<res->size()) { *res+="..."; return plus; }
  plus=printCubes(elsepart,res,start+"-"+ifName,plus,maxlen);
  return plus;
}

void Bdd::Profile(const char* name) const
{
  if(name) printf("%s:\n",name);
  Cal_BddPrintProfile(mgr,rep,BddMgr::printNameVars,NULL,80,stdout);
}

void Bdd::PrintInfo(const char* name) const
{
  if(name) printf("%s: ",name);
  int supplen=0;
  Cal_Bdd *support=new Cal_Bdd[Cal_BddVars(mgr)+1];
  Cal_BddSupport(mgr,rep,support);
  while(support[supplen]) supplen++;
  delete[] support;
  printf("Support %d vars, %ld Nodes, %g satisfying percentage\n",
         supplen,
         Cal_BddSize(mgr,rep,1),
         Cal_BddSatisfyingFraction(mgr,rep));
}

long Bdd::Size() const
{
  return Cal_BddSize(mgr,rep,1);
}

int Bdd::IfIndex() const
{
  return Cal_BddGetIfIndex(mgr,rep);
}

int Bdd::IfId() const
{
  return Cal_BddGetIfId(mgr,rep);
}

Bdd Bdd::IfPart() const
{
  return Cal_BddIf(mgr,rep);
}

Bdd Bdd::ThenPart() const
{
  return Cal_BddThen(mgr,rep);
}

Bdd Bdd::ElsePart() const
{
  return Cal_BddElse(mgr,rep);
}

int Bdd::TotCmp(const Bdd b) const
  // an arbitrary total order on Bdds for use with AVLMaps etc.
{
  return ( (Cal_Bdd(rep) == Cal_Bdd(b.rep)) ? 0 : (Cal_Bdd(rep) < Cal_Bdd(b.rep)) ? -1 : 1 );
}

// int Bdd::Hash() const
// {
//   return multiplicativehash(int(Cal_Bdd(rep)));
// }

ostream& operator << (ostream& s, Bdd b) 
{
  string descr;
  if (100>b.Size() && 800>(descr=b.Describe()).size()){
    s << descr << "\n";
    return s;
  }
  int supplen=0;
  Cal_Bdd *support=new Cal_Bdd[Cal_BddVars(mgr)+1];
  Cal_BddSupport(mgr,b.rep,support);
  while(support[supplen]) supplen++;
  s << supplen << " vars, "
    << Cal_BddSize(mgr,b.rep,1) << " Nodes, "
    << Cal_BddSatisfyingFraction(mgr,b.rep) << " satisfying percentage\n";
//   s << "Support: ";
//   for(int i=0;support[i];i++)
//     s << Bdd(support[i]).VarName() << " ";
//   s << ".\n";
  delete[] support;
  return s;
}

void    Bdd::p() {           // this exists only so I can call it from gdb.
  cout << *this << flush;
}

// ================================================================

BddMgr::BddMgr() : mAssocsValid(FALSE), mForcedOverflowFlag(FALSE)
{
  assert(this==&mgr);           // there is only one BddMgr: mgr.
  theManager=Cal_BddManagerInit();
  assert(Cal_BddManager(mgr));
  Cal_BddDynamicReordering(mgr,CAL_REORDER_NONE);
  Cal_BddManager(mgr)->dynamicReorderingEnableFlag=0;
}

BddMgr::~BddMgr()
{
  if (gPrintWarnings) {
    FILE *fstats=fopen("stats","wt");
    if (NULL!=fstats) {
      Cal_BddStats(mgr,fstats);
      fprintf(fstats,"================================================================\n");
      Cal_BddManagerGC(mgr);
      Cal_BddStats(mgr,fstats);

      BddVector indexSorted;
      { StringunsignedMap::const_iterator it=varNameIdMap.begin(),end=varNameIdMap.end();
      for (;end!=it;++it) indexSorted.push_back(mgr.Var(it->first.c_str())); }
      sort(indexSorted.begin(),indexSorted.end(),Bdd::BddIndexCompare());
      // sort(indexSorted.begin(),indexSorted.end(),Bdd::BddIdCompare());
      { BddVector::const_iterator it=indexSorted.begin(),end=indexSorted.end();
      for (;end!=it;++it) fprintf(fstats,"%s ",it->VarName().c_str()); }
      fprintf(fstats,"\n");
      fclose(fstats);
    }
  }
  Cal_BddManagerQuit(mgr);
  theManager=NULL;              // so that Operative() returns FALSE.
}

// ================================================================
// Variable name related

string BddMgr::varName(Cal_Bdd b)
{
  string res=varIdNameMap[Cal_BddGetIfId(mgr,b)];
  assert(res!="");
  return res;
}

Bdd     BddMgr::Var(const string name)
  // get/create Bdd Variable with name name
  // variables with (back)primes (`) are in the order immediately after
  // variables with no primes.
  // Completely new variables are created at the end of the order.
{
  Cal_BddId_t   id;
  if((id=varNameIdMap[name])!=0) return Cal_BddManagerGetVarWithId(mgr,id);
  // so we create a new variable
  // First, we find least existing variable with less primes, or completely
  // remove primes from name to namenop(rimes)
  mAssocsValid=FALSE;            // assocs have to be recalculated
  assert(name.npos==name.find("'")); // catch accidential use of ' instead of `
  string namenop=name;
  Bdd var,varnop;
  while( ((id=varNameIdMap[namenop])==0) && (lastchar(namenop)=='`')  )
    namenop.erase(namenop.end()-1);
  if(id==0) {                   // namenop doesn't exist yet
    assert(lastchar(namenop)!='`');
    varnop.rep=Cal_BddManagerCreateNewVarLast(mgr);
    insertVarInMaps(varnop,namenop);
  } else {
    varnop.rep=Cal_BddManagerGetVarWithId(mgr,id);
  }
  var.rep=Cal_BddManagerCreateNewVarAfter(mgr,varnop);
  insertVarInMaps(var,name);
  assert(var.IsValid());
  return var;
}

BddVector BddMgr::NPrimedVars(int np)
{
  BddVector res;
  StringunsignedMap &map=mgr.varNameIdMap;
  string primes(np,'`');    
  for (StringunsignedMap::iterator i = map.begin();
       map.end() != i; ++i)
    if(lastchar((*i).first)!='`')
      res.push_back(mgr.Var((*i).first+primes));
  return res;
}

void    BddMgr::insertVarInMaps(Bdd var, string name)
{
  int id=Cal_BddGetIfId(mgr,var);
  varNameIdMap[name]=id;
  varIdNameMap[id]=name;
}

void    BddMgr::updateAssocs()
{
  noprimed.SetToPrimedSet(0);
  primetonoprime.SetToPrimedSubst(1,0);
  primed.SetToPrimedSet(1);
  noprimetoprime.SetToPrimedSubst(0,1);
  mAssocsValid=TRUE;
}

// ================================================================
// Constants etc.

Bdd BddMgr::One() const
{
  return Cal_BddOne(mgr);
}

Bdd BddMgr::Zero() const
{
  return Cal_BddZero(mgr);
}

Bdd BddMgr::Random(double satisFrac, BddVector support)
  // generate Bdd from disjunction of numCubes random cubes.
{
  Bdd res=mgr.Zero(),resold;
  if (0==support.size()) support=mgr.NPrimedVars(0); // the default
  while(Cal_BddSatisfyingFraction(mgr,res)<satisFrac){
    BddVector cube;
    for(BddVector::iterator it=support.begin(); support.end()!=it; ++it)
      if(rndBool()) {
        if(rndBool())   cube.push_back(*it);
        else            cube.push_back(-*it);
      }
    resold=res;
    res+=MultiwayAnd(cube);
  }
  if( Cal_BddSatisfyingFraction(mgr,res)-satisFrac >
      satisFrac-Cal_BddSatisfyingFraction(mgr,resold) )
    return resold;
  else return res;
}

Bdd BddMgr::XorRandom(BddVector support, int ncubes)
  // generate random Bdd from xor of ncubes random cubes
  // support defaults to mgr.NPrimedVars(0), ncubes to support.size()+10
{
  Bdd res;
  if(rndBool()) res=mgr.Zero(); else res=mgr.One();
  if (0==support.size()) support=mgr.NPrimedVars(0); // the default
  if (0==ncubes) ncubes=support.size()+10;
  for(int i=0;ncubes>i;++i) {
    BddVector cube;
    for(BddVector::iterator it=support.begin(); support.end()!=it; ++it)
      if(rndBool()) {
        if(rndBool())   cube.push_back(*it);
        else            cube.push_back(-*it);
      }
    res^=MultiwayAnd(cube);
  }
  return res;
}

// ________________________________________________________________
// statistics

long BddMgr::PeakNumNodes() {    // maximal number of nodes that has been allocated.
  long maxnodes=Cal_BddManager(mgr)->numPeakNodes;
  if(maxnodes<int(Cal_BddManager(mgr)->numNodes))
    maxnodes=Cal_BddManager(mgr)->numNodes;
  return maxnodes;
} 

long BddMgr::GetNumNodes() {    // number of nodes that is allocated.
  return Cal_BddManagerGetNumNodes(mgr);
} 

long BddMgr::SetNodeLimit(long newlimit)
  // sets max. number of bdd nodes, returns old one.
{
  return Cal_BddNodeLimit(mgr,newlimit);
}

void BddMgr::EnableGC(bool dogc)
  // enables / disables GC
{
  Cal_BddSetGCMode(mgr, dogc ? 1 : 0 );
}

void BddMgr::DoGC()
  // does a garbage collection
{
  Cal_BddManagerGC(mgr);
}

bool BddMgr::WasOverflow()
  // sets max. number of bdd nodes, returns old one.
{
  bool res=Cal_BddOverflow(mgr) || mForcedOverflowFlag;
  mForcedOverflowFlag=FALSE;
  return res;
}

void BddMgr::SetOverflowFlag()
{ // forces us to behave like there was an overflow.
  mForcedOverflowFlag=TRUE;
}

void BddMgr::Reorder(int method) 
{ // invokes reordering
  Cal_BddDynamicReordering(mgr,method);
  Cal_BddReorder(mgr);
  Cal_BddDynamicReordering(mgr,CAL_REORDER_NONE);
} 

void BddMgr::SetDynReordering(int method)
{ // allows/disallows dynamic reordering
    Cal_BddDynamicReordering(mgr,method);
  if (CAL_REORDER_NONE==method) {
    Cal_BddManager(mgr)->dynamicReorderingEnableFlag=0;
  } else {
    Cal_BddManager(mgr)->dynamicReorderingEnableFlag=1;
  }
}

// ================================================================
// Association Implementation

BddAssoc::BddAssoc() : assocId(-1), onlyVars(FALSE), isMap(FALSE)
{
}

BddAssoc::BddAssoc(const BddVector& set)
{
  assign(set);
}

BddAssoc& BddAssoc::operator=(const BddVector& set)
{
  unassign();
  assign(set);
  return *this;
}

void BddAssoc::assign(const BddVector& set)
{
  onlyVars=TRUE;
  isMap=FALSE;
  int len=set.size();
  Cal_Bdd *assInfo=new Cal_Bdd[len+1];
  Cal_Bdd *next=assInfo;
  for (BddVector::const_iterator i = set.begin(); i != set.end(); i++, next++) {
    assert(i->IsVariable());
    *next=*i;
  }
  *next=NULL;
  assocId=Cal_AssociationInit(mgr,assInfo,0);
  assert(assocId!=-1);
  delete[] assInfo;
}

BddAssoc::BddAssoc(const BddBddMap& map)
{
  assign(map);
}

BddAssoc& BddAssoc::operator=(const BddBddMap& map)
{
  unassign();
  assign(map);
  return *this;
}

BddAssoc::BddAssoc(const Bdd b)
{
  assert(b.IsVariable());
  Cal_Bdd assInfo[2]={b.rep,NULL};
  assocId=Cal_AssociationInit(mgr,assInfo,0);
  onlyVars=b.IsVariable();
  isMap=FALSE;
  assert(assocId!=-1);
}

void BddAssoc::assign(const BddBddMap& map)
{
  onlyVars=TRUE;
  isMap=TRUE;
  int len=map.size();
  Cal_Bdd *assInfo=new Cal_Bdd[2*len+2];
  Cal_Bdd *next=assInfo;
  for (BddBddMap::const_iterator i = map.begin(); i != map.end(); i++, ++ ++ next){
    *(next++)=(*i).first; assert(i->first.IsVariable());
    *(next++)=(*i).second;
    if(onlyVars && -i->second.IsVariable()) onlyVars=FALSE;
  }
  *next++=NULL; *next++=NULL;
  assocId=Cal_AssociationInit(mgr,assInfo,1);
  assert(assocId!=-1);
  delete[] assInfo;
}

BddAssoc::~BddAssoc()
{
  if (!mgr.Operative()) return;
  unassign();
}

void BddAssoc::unassign()
{
  if(assocId!=-1) Cal_AssociationQuit(mgr,assocId);
  assocId=-1;
  onlyVars=FALSE;
}

int BddAssoc::Id() const
{
  return assocId;
}

bool  BddAssoc::IsMapping() const { // true if assoc is not an array but a mapping
  return isMap;
}

bool  BddAssoc::MapsToVariables() const { // true if assoc maps only to variables
  assert(IsMapping());
  return onlyVars;
}

void BddAssoc::SetToPrimedSet(int np)
{
  StringunsignedMap &map=mgr.varNameIdMap;
  int len=map.size();
  Cal_Bdd *assInfo=new Cal_Bdd[len+1];
  Cal_Bdd *next=assInfo;
  string primes(np,'`');
  for (StringunsignedMap::iterator i = map.begin();
       i != map.end(); i++)
    if(lastchar((*i).first)!='`')
      *(next++)=mgr.Var((*i).first+primes);
  *next=NULL;
  assocId=Cal_AssociationInit(mgr,assInfo,0);
  assert(assocId!=-1);
  onlyVars=TRUE;
  isMap=FALSE;
  delete[] assInfo;
}

void BddAssoc::SetToPrimedSubst(int np1, int np2)
{
  StringunsignedMap &map=mgr.varNameIdMap;
  int len=map.size();
  Cal_Bdd *assInfo=new Cal_Bdd[2*len+2];
  Cal_Bdd *next=assInfo;
  string prim1(np1,'`');
  string prim2(np2,'`');
  for (StringunsignedMap::iterator i = map.begin(); i != map.end();
       ++i)
    if(lastchar((*i).first)!='`') {
      *(next++)=mgr.Var((*i).first+prim1);
      *(next++)=mgr.Var((*i).first+prim2);
    }
  *next++=NULL; *next++=NULL;
  assocId=Cal_AssociationInit(mgr,assInfo,1);
  assert(assocId!=-1);
  onlyVars=TRUE;
  isMap=TRUE;
  delete[] assInfo;
}

void BddAssoc::SetToSwitchSubst(int np1, int np2)
{
  StringunsignedMap &map=mgr.varNameIdMap;
  int len=map.size();
  Cal_Bdd *assInfo=new Cal_Bdd[4*len+2];
  Cal_Bdd *next=assInfo;
  string prim1(np1,'`');
  string prim2(np2,'`');
  for (StringunsignedMap::iterator i = map.begin(); i != map.end();
       ++i)
    if(lastchar((*i).first)!='`') {
      *(next++)=mgr.Var((*i).first+prim1);
      *(next++)=mgr.Var((*i).first+prim2);
    }
  for (StringunsignedMap::iterator i = map.begin(); i != map.end();
       ++i)
    if(lastchar((*i).first)!='`') {
      *(next++)=mgr.Var((*i).first+prim2);
      *(next++)=mgr.Var((*i).first+prim1);
    }
  *next++=NULL; *next++=NULL;
  assocId=Cal_AssociationInit(mgr,assInfo,1);
  assert(assocId!=-1);
  onlyVars=TRUE;
  isMap=TRUE;
  delete[] assInfo;
}

